import numpy as np
import matplotlib.pyplot as plt
from .marching_cubes import ger_iso_meshs, mesh_concatenate
from ..colormap import Colormap
from ..mesh import Mesh

def get_iso_surf(res, **kwargs):
    
    if res.min() < 0:
        cmap = kwargs.pop("cmap", plt.cm.seismic)
        vmax = kwargs.pop("vmax", min(np.abs(res.max()), np.abs(res.min())) * 1.0)
        vmin = kwargs.pop("vmin", -vmax)
    else:
        cmap = kwargs.pop("cmap", plt.cm.jet)
        vmax = kwargs.pop("vmax", res.max() * 0.85)
        vmin = kwargs.pop("vmin", res.max() * 0.15)
    
    colormap = Colormap(plt.cm.jet, vmin, vmax)
    contours_number = kwargs.pop("contours_number", 4)
    contours = kwargs.pop("contours", np.linspace(vmin, vmax, contours_number))
    
    tri_verts, tri_faces, tri_norms, tri_color, iso_vals = ger_iso_meshs(res, contours, colormap)
    verts, faces, normals, colors = mesh_concatenate(tri_verts, tri_faces, tri_norms, tri_color)
    return Mesh(verts, faces, normals, colors, iso_vals)


import numpy as np
import trimesh
from trimesh import Trimesh
from ..show import plt_trisurf, Poly3D

class Mesh(Trimesh):
    
    def __init__(self, verts, faces, normals, colors, iso_vals):
        
        self.iso_vals = iso_vals
        self.mesh = trimesh.Trimesh(vertices = verts, faces = faces,
                           vertex_normals = normals, vertex_colors = colors)
        self.verts = self.mesh.vertices
        self.faces = self.mesh.faces
        self.normals = self.mesh.vertex_normals
        self.colors = self.mesh.visual.vertex_colors
        return
    
    def load(filename, ext = "obj"):
        mesh = trimesh.load(f"{filename}.{ext}", ext)
        iso_vals = np.load(f"{filename}.npy")
        return Mesh(mesh.vertices, mesh.faces, mesh.vertex_normals, mesh.visual.vertex_colors, iso_vals)
    
    def export(self, filename, ext = "obj"):
        self.mesh.export(f"{filename}.{ext}", ext)
        np.save(f"{filename}.npy", self.iso_vals)
        return
    
    def plt_trisurf(self, **kwargs):
        verts, faces, iso_vals = self.verts, self.faces, self.iso_vals
        surf = plt_trisurf(verts, faces, array = iso_vals, **kwargs)
        return surf
    
    def Poly3D(self, **kwargs):
        verts, faces, iso_vals = self.verts, self.faces, self.iso_vals
        surf = Poly3D(verts, faces, array = iso_vals, **kwargs)
        return surf


def iso_surface(mesh, **kwargs):
    from mayavi import mlab
    mlab_mesh = mlab.triangular_mesh(mesh.verts[:, 0], mesh.verts[:, 1], mesh.verts[:, 2], 
                                    mesh.faces, scalars = mesh.iso_vals, **kwargs)
    return mlab_mesh
    


import trimesh
import numpy as np
from skimage import measure
import matplotlib.pyplot as plt

def marching_cubes(res, iso_value, colormap):
    verts, faces, normals, values = measure.marching_cubes(res, iso_value, spacing = [1.0, 1.0, 1.0])

    pltmap = colormap.cmap
        
    vmax = colormap.vmax
    vmin = colormap.vmin
    
    #vmax = ey.max()
    #vmin = -vmax

    xp = [vmin, vmax]
    fp = [0, 256]

    color = pltmap(int(np.interp(iso_value, xp, fp)))
    colors = np.asarray([color for _ in range(verts.shape[0])])
    #color = pltmap(np.asarray(np.interp(values, xp, fp), dtype = "int"))
    return verts, faces, normals, colors

def ger_iso_meshs(res, iso_value_range, colormap):

    tri_verts = []
    tri_faces = []
    tri_norms = []
    tri_color = []
    #color = []
    iso_vals = np.zeros([0])

    for iso_value in iso_value_range:
        verts, faces, normals, colors = marching_cubes(res, iso_value, colormap)

        tri_verts.append(verts)
        tri_faces.append(faces)
        tri_norms.append(normals)
        tri_color.append(colors)

        #color.append(colors[0])
        iso_vals = np.hstack((iso_vals, np.ones_like(verts[:, 0]) * iso_value))
        
    return tri_verts, tri_faces, tri_norms, tri_color, np.asarray(iso_vals)

def mesh_concatenate(tri_verts, tri_faces, tri_norms, tri_color):
    v_old = np.zeros([0, 3])
    f_old = np.zeros([0, 3], dtype = "int")
    n_old = np.zeros([0, 3])
    c_old = np.zeros([0, 4], dtype = "int")

    for i, verts in enumerate(tri_verts):

        #matching
        f_new = np.array(tri_faces[i]) + np.shape(v_old)[0]
        v_old = np.concatenate((v_old,tri_verts[i]),axis=0)
        f_old = np.concatenate((f_old,f_new),axis=0)
        n_old = np.concatenate((n_old,tri_norms[i]),axis=0)
        c_old = np.concatenate((c_old,tri_color[i]),axis=0)
        
    return v_old, f_old, n_old, c_old


import numpy as np
import matplotlib.pyplot as plt
from numba import njit, prange

@njit(parallel = True)
def interpolate_3d(ey, res):
    m, n = ey.shape[0], ey.shape[1]
    y0 = (n - 1)/2
    z0 = (n - 1)/2
    for i in prange(m):
        for j in range(n):
            for k in range(n):
                r = int(np.sqrt((j - y0)**2 + (k - y0)**2))
                if int(y0 + r) > n - 1 or int(y0 - r) < 0:
                    continue
                z = k - z0
                y = j - y0
                theta = np.arctan2(z, y)
                if np.abs(theta) <= 3.14/2:
                    res[i, j, k] = ey[i, int(y0 + r)]
                #else:
                #    res[i, j, k] = ey[i, int(y0 - r)]
            
    return

@njit(parallel = True)
def rotate_total(ey, res):
    m, n = ey.shape[0], ey.shape[1]
    y0 = (n - 1)/2
    z0 = (n - 1)/2
    for i in prange(m):
        for j in range(n):
            for k in range(n):
                r = int(np.sqrt((j - y0)**2 + (k - y0)**2))
                if int(r + y0) < n:
                    res[i, j, k] = ey[i, int(r + y0)]
    
    return

def rotate(ey, res, ifhalf = False):
    if ifhalf:
        interpolate_3d(ey, res)
    else:
        rotate_total(ey, res)
    return